/*
 * keytree.c
 *
 * - KEYTREE held by applications
 */

#include "seal.h"

/*** declarations of the linked list structure ***/
typedef struct keytree_list_node_s * KEYTREE_list_node_p_t;
typedef struct keytree_list_node_s {
	KEYTREE_tree_node_p_t tnode;
	KEYTREE_list_node_p_t prev;
	KEYTREE_list_node_p_t next;
} KEYTREE_list_node_t;

typedef struct keytree_list_s * KEYTREE_list_p_t;
typedef struct keytree_list_s {
	KEYTREE_list_node_p_t head;
	KEYTREE_list_node_p_t tail;
} KEYTREE_list_t;

/*** basic operations on the keytree ***/
static void KEYTREE_single_join(KEYTREE_control_p_t, char const*);
static void KEYTREE_single_leave(KEYTREE_control_p_t, char const*);
static void KEYTREE_pure_join(KEYTREE_control_p_t, char* const* joins, 
		int join_size, KEYTREE_tree_node_p_t insertion);
static void KEYTREE_pure_leave(KEYTREE_control_p_t, char* const* leaves, 
		int leave_size);

/*** linked list handling ***/
static KEYTREE_list_p_t KEYTREE_list_init();
static void KEYTREE_list_addq(KEYTREE_list_p_t, KEYTREE_tree_node_p_t);
static KEYTREE_tree_node_p_t KEYTREE_list_delq(KEYTREE_list_p_t);
static void KEYTREE_list_addi(KEYTREE_list_p_t, KEYTREE_tree_node_p_t);
static KEYTREE_tree_node_p_t KEYTREE_list_pop(KEYTREE_list_p_t);
static void KEYTREE_list_destroy(KEYTREE_list_p_t);

/*** helper functions ***/
static void KEYTREE_update_tree_node_id(KEYTREE_tree_node_p_t, int root_id);
static void KEYTREE_mark_renewed(KEYTREE_tree_node_p_t);
static void KEYTREE_unmark_renewed(KEYTREE_tree_node_p_t);
static void KEYTREE_mark_sponsor(KEYTREE_tree_node_p_t);
static void KEYTREE_unmark_sponsor(KEYTREE_control_p_t);
static void KEYTREE_mark_all_sponsor(KEYTREE_tree_node_p_t);
static int KEYTREE_get_height(KEYTREE_tree_node_p_t);

/*** log2 ***/
static int log_2(int x);

/*---------------------------------------------------------------------
 *				KEYTREE initialization
 *--------------------------------------------------------------------*/
/*
 * Initialize the keytree structure
 */
KEYTREE_control_p_t KEYTREE_init() {
	KEYTREE_control_p_t tree;

	tree = (KEYTREE_control_p_t)malloc(sizeof(KEYTREE_control_t));
	tree->root = NULL;
	tree->m_list_head = NULL;
	tree->m_list_tail = NULL;

	tree->subtree = (KEYTREE_control_p_t)malloc(sizeof(KEYTREE_control_t));
	tree->subtree->root = NULL;
	tree->subtree->m_list_head = NULL;
	tree->subtree->m_list_tail = NULL;
	tree->subtree->subtree = NULL;			// never used

	pthread_mutex_init(&(tree->lock), NULL);

	return tree;
}

/*
 * Create a keytree given the num of members and root ids
 */
void KEYTREE_create_from_size(KEYTREE_control_p_t tree, char* const* members, 
		int size) {
	KEYTREE_tree_node_p_t* k_nodes;
	int i;
	
	// initialize all k-nodes, there should be 2*size-1 k-nodes 
	k_nodes = (KEYTREE_tree_node_p_t*)malloc((2 * size - 1) *
			sizeof(KEYTREE_tree_node_p_t));
	for (i=0; i<2*size-1; ++i) {
		k_nodes[i] = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
		k_nodes[i]->left = NULL;
		k_nodes[i]->right = NULL;
		k_nodes[i]->parent = NULL;
		k_nodes[i]->mem_node = NULL;
		k_nodes[i]->tree_node_id = -1;
		k_nodes[i]->renewed = 1;
	}
	
	// 
	// create a tree of all k-nodes, in BFS order 
	//
	for (i=0; i<2*size-1; ++i) {
		if (i == 0) {	// update the tree root 
			tree->root = k_nodes[0];	
			k_nodes[0]->tree_node_id = 0;
		} else {
			k_nodes[i]->parent = k_nodes[(i-1)/2];
			if (i % 2 == 1) {	 // tree nodes with odd tree_node_id 
				k_nodes[i]->parent->left = k_nodes[i];
				k_nodes[i]->tree_node_id = 2 * k_nodes[i]->parent->tree_node_id
										+ 1;
			} else {	 // tree nodes with even tree_node_id 
				k_nodes[i]->parent->right = k_nodes[i];
				k_nodes[i]->tree_node_id = 2 * k_nodes[i]->parent->tree_node_id
										+ 2;
			}
		}
	}
	
	// create m-nodes under the last "size" k-nodes 
	for (i=size-1; i<2*size-1; ++i) {
		k_nodes[i]->mem_node = (KEYTREE_mem_node_p_t)
			malloc(sizeof(KEYTREE_mem_node_t));
		memcpy(k_nodes[i]->mem_node->mem_guid, members[i-size+1], 
				SEAL_GUID_LEN);
		k_nodes[i]->mem_node->tree_node = k_nodes[i];
		k_nodes[i]->mem_node->sponsor = 0;

		k_nodes[i]->renewed = 0;		// reset the leaf_node->renewed = 0

		// add to user list 
		if (i == size -1) {
			tree->m_list_head = k_nodes[i]->mem_node;
			tree->m_list_tail = k_nodes[i]->mem_node;
			k_nodes[i]->mem_node->prev = NULL;
		} else {
			k_nodes[i]->mem_node->prev = tree->m_list_tail;
			tree->m_list_tail->next = k_nodes[i]->mem_node;
			tree->m_list_tail = k_nodes[i]->mem_node;
		}
		k_nodes[i]->mem_node->next = NULL;
	}
}

/*
 * Create a keytree given the member guids and their tree node ids
 */
void KEYTREE_create_from_list(KEYTREE_control_p_t tree, char* const* members,
		int const* tree_node_ids, int size) {

	KEYTREE_tree_node_p_t k_node;
	KEYTREE_tree_node_p_t tmp_k_node1, tmp_k_node2;
	KEYTREE_list_p_t list;
	int i;

	// create all leave nodes associating the users
	list = KEYTREE_list_init();
	for (i=0; i<size; ++i) {
		k_node = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
		k_node->left = NULL;
		k_node->right = NULL;
		k_node->parent = NULL;
		k_node->tree_node_id = tree_node_ids[i];
		k_node->renewed = 0;
		k_node->mem_node = (KEYTREE_mem_node_p_t)
			malloc(sizeof(KEYTREE_mem_node_t));
		memcpy(k_node->mem_node->mem_guid, members[i], SEAL_GUID_LEN);
		k_node->mem_node->tree_node = k_node;
		k_node->mem_node->sponsor = 0;
		if (tree->m_list_head == NULL) {
			k_node->mem_node->prev = NULL;
			k_node->mem_node->next = NULL;
			tree->m_list_head = k_node->mem_node;
			tree->m_list_tail = k_node->mem_node;
		} else {
			k_node->mem_node->prev = tree->m_list_tail;
			k_node->mem_node->next = NULL;
			tree->m_list_tail->next = k_node->mem_node;
			tree->m_list_tail = k_node->mem_node;
		}
		// add to the temp list in increasing order of tree_node_id
		KEYTREE_list_addi(list, k_node);
	}
	
	// expand until the root node is reached
	tmp_k_node1 = KEYTREE_list_pop(list);
	tmp_k_node2 = KEYTREE_list_pop(list);
	while (tmp_k_node2 != NULL) {
		// create a parent node	
		if (tmp_k_node1->tree_node_id - tmp_k_node2->tree_node_id != 1) {
			debug_err("unmatched tree_node_id in creation\n");
			return;
		}
		k_node = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
		k_node->right = tmp_k_node1;
		k_node->left = tmp_k_node2;
		k_node->parent = NULL;
		tmp_k_node1->parent = k_node;
		tmp_k_node2->parent = k_node;
		k_node->tree_node_id = tmp_k_node2->tree_node_id / 2;
		k_node->renewed = 0;
		k_node->mem_node = NULL;
		KEYTREE_list_addi(list, k_node);
		tmp_k_node1 = KEYTREE_list_pop(list);
		tmp_k_node2 = KEYTREE_list_pop(list);
	}
	tree->root = tmp_k_node1;

	// free all resources
	KEYTREE_list_destroy(list);
}


/*----------------------------------------------------------------------
*                            REBUILD                         
*-----------------------------------------------------------------------*/
/*
 * Rebuild: Create a balanced key tree in BFS order
 */
void KEYTREE_rebuild(KEYTREE_control_p_t t, char* const* joins, int join_size,
		char* const* leaves, int leave_size) {

	char** memlist, **new_memlist;				// the set of members
	int mem_size, new_mem_size;					// size of memlist
	KEYTREE_mem_node_p_t m_node;
	KEYTREE_control_p_t tmp_t;
	char tmp_member[SEAL_GUID_LEN];	
	int i, j;
	
	// obtain all memlist from the current tree
	mem_size = KEYTREE_get_num_of_members(t);
	memlist = (char**)malloc(mem_size * sizeof(char*));	
	m_node = t->m_list_head;
	i = 0;
	while (m_node != NULL) {
		memlist[i] = (char*)malloc(SEAL_GUID_LEN);
		memcpy(memlist[i], m_node->mem_guid, SEAL_GUID_LEN);
		i++;
		m_node = m_node->next;
	}

	// remove the L leaving memlist from memlist
	for (i=0; i<leave_size; ++i) {
		for (j=0; j<mem_size-i; ++j) {
			if (memcmp(leaves[i], memlist[j], SEAL_GUID_LEN) == 0) {
				memcpy(memlist[j], memlist[mem_size-i-1], SEAL_GUID_LEN);
				break;
			}	
		}
		if (j == mem_size - i) {
			debug_err("member %s not found in reb\n", leaves[i]);
			return;
		}
	}

	// add the J new memlist to new memlist
	new_mem_size = mem_size - leave_size + join_size;
	new_memlist = (char**)malloc(new_mem_size * sizeof(char*));

	for (i=0; i<mem_size-leave_size; ++i) {
		new_memlist[i] = (char*)malloc(SEAL_GUID_LEN);
		memcpy(new_memlist[i], memlist[i], SEAL_GUID_LEN);
	}

	for (i=0; i<join_size; ++i) {
		new_memlist[i + mem_size - leave_size] = (char*)malloc(SEAL_GUID_LEN);
		memcpy(new_memlist[i + mem_size - leave_size], joins[i], SEAL_GUID_LEN);
	}

	// sort the memlist
	for (i=0; i<new_mem_size-1; ++i) {
		for (j=0; j<new_mem_size-1-i; ++j) {
			if (memcmp(new_memlist[j], new_memlist[j+1], SEAL_GUID_LEN) > 0) {
				memcpy(tmp_member, new_memlist[j], SEAL_GUID_LEN);
				memcpy(new_memlist[j], new_memlist[j+1], SEAL_GUID_LEN);
				memcpy(new_memlist[j+1], tmp_member, SEAL_GUID_LEN);
			}
		}
	}

	// rebuild the new binary tree 
	tmp_t = KEYTREE_init();					// create a tmp tree
	tmp_t->root = t->root;					// copy t to tmp_t
	tmp_t->m_list_head = t->m_list_head;
	tmp_t->m_list_tail = t->m_list_tail;
	KEYTREE_destroy(tmp_t);
	tmp_t = KEYTREE_init();
	KEYTREE_create_from_size(tmp_t, new_memlist, new_mem_size);
	t->root = tmp_t->root;					// copy tmp_t back to t
	t->m_list_head = tmp_t->m_list_head;
	t->m_list_tail = tmp_t->m_list_tail;

	// mark all sponsors	
	KEYTREE_mark_all_sponsor(t->root);
	
	// free resources
	for (i=0; i<mem_size; ++i) {
		free(memlist[i]);
	}
	for (i=0; i<new_mem_size; ++i) {
		free(new_memlist[i]);
	}
	free(memlist);
	free(new_memlist);
	free(tmp_t);
}	
	
/*----------------------------------------------------------------------
*                            BATCH                         
*-----------------------------------------------------------------------*/
/*
 * The main Batch algorithm
 */
void KEYTREE_batch(KEYTREE_control_p_t tree, char* const* joins, int join_size,
		char* const* leaves, int leave_size) {

	KEYTREE_mem_node_p_t* leave_m_nodes;	// ptrs to leaving member nodes
	KEYTREE_mem_node_p_t tmp_m_node;		// temporary member node
	char tmp_mem_guid[SEAL_GUID_LEN];			// temporary member guid
	int subgroup_size;				// size of subgroups (used when J > L > 0)	
	int i, j;

	if (leave_size == 0) {		// pure join case

		// create a subtree for the new joins, and
		// attach the subtree to the shallowest position
		KEYTREE_pure_join(tree, joins, join_size, NULL);	

	} else {			// leave size > 0
		
		// Get the leave member list in the ascending order or tree node id
		// i) malloc
		leave_m_nodes = (KEYTREE_mem_node_p_t*)malloc(leave_size * 
				sizeof(KEYTREE_mem_node_p_t));
		// ii) get informatio for all leave_mem_nodes
		for (i=0; i<leave_size; ++i) {
			tmp_m_node = tree->m_list_head;
			while (tmp_m_node != NULL) {
				if (memcmp(tmp_m_node->mem_guid, leaves[i], 
							SEAL_GUID_LEN) == 0) {
					leave_m_nodes[i] = tmp_m_node;
					break;
				}
				tmp_m_node = tmp_m_node->next;
			}
			if (tmp_m_node == NULL) {
				debug_err("member %s not found in batch\n", leaves[i]);
				exit(-1);
			}	
		}
		// iii) sort
		for (i=0; i<leave_size; ++i) {
			for (j=0; j<leave_size-1-i; ++j) {
				if (leave_m_nodes[j]->tree_node->tree_node_id 
						> leave_m_nodes[j+1]->tree_node->tree_node_id) {
					tmp_m_node = leave_m_nodes[j];
					leave_m_nodes[j] = leave_m_nodes[j+1];
					leave_m_nodes[j+1] = tmp_m_node;	
				}
			}
		}	// end: sorting
	
		if (leave_size >= join_size) { // more members want to leave than join
			
			if (join_size > 0) {
				// replace the departed nodes with join nodes
				for (i=0; i<join_size; ++i) {
					memcpy(leave_m_nodes[i]->mem_guid, joins[i], SEAL_GUID_LEN);
					KEYTREE_mark_renewed(
							leave_m_nodes[i]->tree_node->parent);
				}
			}

			if (leave_size - join_size > 0) {
				// remove the remaining L - J nodes, and 
				// promote the siblings
				for (i=join_size; i<leave_size; ++i) {
					memcpy(tmp_mem_guid, leave_m_nodes[i]->mem_guid, 
							SEAL_GUID_LEN);
					KEYTREE_single_leave(tree, tmp_mem_guid);
				}
			}

		} else {	// more newly joining members than leaving members
		
			// (i) divide joins into L subgroups, and
			// (ii) create L subtrees for the subgroups	
			// (iii) replace the departed nodes with subtrees' roots
			j = 0;				// count how many joins are added
			for (i=0; i<leave_size; ++i) {
				subgroup_size = i < (join_size % leave_size) ? 
					join_size / leave_size + 1 : join_size / leave_size;
			
				// (ii) and (iii) equivalent to: first do pure join then do 
				// pure leave	
				KEYTREE_pure_join(tree, joins + j, subgroup_size, 
						leave_m_nodes[i]->tree_node);
				KEYTREE_single_leave(tree, leave_m_nodes[i]->mem_guid);		

				j += subgroup_size;
			}

		}	// end: more newly joining members than leaving members
		
		// free resources
		free(leave_m_nodes);

	}	// end: if L > 0

	// elect all sponsors
	KEYTREE_mark_all_sponsor(tree->root);
}

/*------------------------------------------------------------------------
* QUEUE-BATCH
*
* First create a subtree on the potentially joined users. Then merge the
* tree with the existing tree, either at the highest leaf node or the
* highest leave location.
*-------------------------------------------------------------------------*/
/*
 * Queue-subtree
 */
void KEYTREE_queue_subtree(KEYTREE_control_p_t subtree, char* const* joins,
		int join_size, char* const* leaves, int leave_size) {
	
	// single join
	if (join_size == 1) {
		KEYTREE_single_join(subtree, joins[0]);
	}

	if (leave_size == 1) {
		KEYTREE_single_leave(subtree, leaves[0]);
	}

	// mark required sponsors	
	KEYTREE_mark_all_sponsor(subtree->root);
}

/*
 * Queue-merge
 */
void KEYTREE_queue_merge(KEYTREE_control_p_t tree, char* const* joins, 
		int join_size, char* const* leaves, int leave_size) {

	KEYTREE_control_p_t subtree;
	KEYTREE_tree_node_p_t k_ins_node, k_node;
	KEYTREE_list_p_t list;
	int is_ins_found = 0;
	
	int height_t;       // height of the existing tree 
	int height_j;       // height the subtree of joined users
	int node_height;    // height of the subtree rooted at the node
	int node_level;		// level of a node (root level = 0)

	int min_tree_node_id = INT_MAX;
	KEYTREE_mem_node_p_t m_node;
	int i, j;			// loop counters

	// Assign the subtree
	subtree = tree->subtree;

	// Error checking: ensure non-null tree
	if (tree == NULL) {
		debug_err("Tree is not created\n");
		return;
	}

	// add to subtree if the join does not exist
	for (i=0; i<join_size; ++i) {
		if (!KEYTREE_is_guid_exist(subtree, joins[i])) {
			KEYTREE_single_join(subtree, joins[i]);
		}
	}

	// remove from the subtree if the leave exists
	for (i=0; i<leave_size; ++i) {
		if (KEYTREE_is_guid_exist(subtree, leaves[i])) {
			KEYTREE_single_leave(subtree, leaves[i]);
			for (j=i; j<leave_size-1; ++j) {
				memcpy(leaves[j], leaves[j+1], SEAL_GUID_LEN);
			}
			leave_size--;
			i--;
		}
	}

	if (join_size > 0) {
		// Find the insertion node: k_ins_node 
		//  Find the highest leaf node if no leaves.
		//  Find the highest departed leaf node if leaves occur.
		//  Without specified, use BFS to find the insetion node, 
		//  which is the highest leaf node. 
		if (subtree->root != NULL) {		// non-empty subtree
			if (tree->root == NULL) {
				k_ins_node = NULL;
			} else {
				if (leave_size == 0) {
					height_t = KEYTREE_get_height(tree->root);
					height_j = KEYTREE_get_height(subtree->root);
					list = KEYTREE_list_init();
					KEYTREE_list_addq(list, tree->root);
					k_ins_node = KEYTREE_list_delq(list);
					while (!is_ins_found && k_ins_node != NULL) {
						node_height = KEYTREE_get_height(k_ins_node);
						node_level = log_2(k_ins_node->tree_node_id+1) ;
						if (((node_height > height_j && node_level + 1 + 
								node_height <= height_t) || 
							(height_j >= node_height && node_level + 1 + 
								height_j <= height_t))){
							is_ins_found = 1;
						} else {
							if (k_ins_node->left != NULL
									&& k_ins_node->right != NULL) {
								KEYTREE_list_addq(list, k_ins_node->left);
								KEYTREE_list_addq(list, k_ins_node->right);
							}
							k_ins_node = KEYTREE_list_delq(list);
						}
					}
					if (is_ins_found == 0) {  // inserted at root
						k_ins_node = tree->root;
					}
					KEYTREE_list_destroy(list);
				} else {			// leave_size > 0
					k_ins_node = NULL;	
					for (i=0; i<leave_size; ++i) {
						m_node = tree->m_list_head;
						while (m_node != NULL && memcmp(m_node->mem_guid, 
									leaves[i], SEAL_GUID_LEN) != 0) {
							m_node = m_node->next;
						}
						if (m_node != NULL) {
							if (m_node->tree_node->tree_node_id 
									< min_tree_node_id) {
								min_tree_node_id = 
									m_node->tree_node->tree_node_id;
								k_ins_node = m_node->tree_node;
							}
						} else {
							debug_err("user %s not found\n", leaves[i]);
							return;
						}
					}
				}
			}

			// Merge the tree 
			if (tree->root == NULL) {  // at the tree root (k_ins_node == NULL) 

				// set the tree equal to the subtree 
				tree->root = subtree->root;
				tree->m_list_head = subtree->m_list_head;
				tree->m_list_tail = subtree->m_list_tail;

			} else {			// k_ins_node exists 
				// Create a new node which will replace the insertion node, then
				// the new node has two children: subtree->root and k_ins_node 
				// (different from the operation of pure_join) 

				// Create a new node that replaces the insertion node 
				k_node = (KEYTREE_tree_node_p_t)malloc(sizeof(
							KEYTREE_tree_node_t));
				k_node->parent = k_ins_node->parent;
				k_node->left = subtree->root;
				k_node->right = k_ins_node;
				k_node->mem_node = NULL;
				k_node->tree_node_id = k_ins_node->tree_node_id;
				// Update insertion node's parent (or k_node's parent) 
				if (k_node->parent == NULL) {
					tree->root = k_node;
				} else if (k_node->parent->left == k_ins_node) {
					k_node->parent->left = k_node;
				} else {
					k_node->parent->right = k_node;
				}
				// Update the insertion node 
				k_ins_node->parent = k_node;
				KEYTREE_update_tree_node_id(k_ins_node,
						2 * k_node->tree_node_id + 2);

				// Update the subtree's root 
				subtree->root->parent = k_node;
				KEYTREE_update_tree_node_id(subtree->root,
						2 * k_node->tree_node_id + 1);

				// Mark the tree renewed / sponsor 
				KEYTREE_mark_renewed(subtree->root->parent);

				// add the user list of the subtree to that of
				// the main tree 
				// (users exist in the main tree) 
				tree->m_list_tail->next = subtree->m_list_head;
				subtree->m_list_head->prev = tree->m_list_tail;
				tree->m_list_tail = subtree->m_list_tail;

			} // End of k_ins_node exists 
		}	// End of non-empty subtree
	
		// reset the subtree
		subtree->root = NULL;
		subtree->m_list_head = NULL;
		subtree->m_list_tail = NULL;
			
	} // End of join_size > 0
	
	// do the pure leave, if any
	if (leave_size > 0) {
		KEYTREE_pure_leave(tree, leaves, leave_size);
	}
	
	// mark all sponsors	
	KEYTREE_mark_all_sponsor(tree->root);
}

/*
 * Sponsors' cooridination algorithms
 *
 * Return the number of broadcast keys, and also the node ids to be 
 * broadcast
 */
int KEYTREE_sponsor_coordinate(KEYTREE_control_p_t tree, char const* guid,
		int* bcast_list) {
	KEYTREE_mem_node_p_t m_node;
	KEYTREE_tree_node_p_t k_node, k_sib_node;
	int no_bcast = 0;

	// locate the m_node
	m_node = tree->m_list_head;
	while (m_node != NULL) {
		if (memcmp(m_node->mem_guid, guid, SEAL_GUID_LEN) == 0) {
			break;
		} else {
			m_node = m_node->next;
		}
	}
	if (m_node == NULL) {
		debug_err("member not found in sponsor_coordinate\n");
		KEYTREE_print(tree->root);
		return -1;
	}

	// determine which keys to be broadcast. If sponsor, the parent of the
	// leaf node is non-null.
	if (m_node->sponsor) {
		k_node = m_node->tree_node;
		while (!(k_node->parent->renewed)) {  // parent != NULL if sponsor
			k_node = k_node->parent;
		}
		bcast_list[no_bcast] = k_node->tree_node_id;
		no_bcast++;
		while (k_node->parent != NULL) {
			// determine the sibling node
			if (k_node->tree_node_id % 2 == 1) {	// k_node=left child
				k_sib_node = k_node->parent->right;
			} else {								// k_node=right child
				k_sib_node = k_node->parent->left;
			}

			if (!(k_node->renewed) && !(k_sib_node->renewed)) {
				if (k_node->tree_node_id % 2 == 1) {	// k_node=left child
					break;	
				} else {								// k_node=right child
					bcast_list[no_bcast] = k_node->parent->tree_node_id;
					no_bcast++;
				}
			} else if (!(k_node->renewed) && k_sib_node->renewed) {
				break;
			} else if (k_node->renewed && !(k_sib_node->renewed)) {
				bcast_list[no_bcast] = k_node->parent->tree_node_id;
				no_bcast++;
			} else {
				if (k_node->tree_node_id % 2 == 1) {	// k_node=left child
					break;	
				} else {								// k_node=right child
					bcast_list[no_bcast] = k_node->parent->tree_node_id;
					no_bcast++;
				}
			}
			k_node = k_node->parent;
		}
	}
	return no_bcast;
}


/*------------------------------------------------------------------------
 *                       Helper Functions
 *------------------------------------------------------------------------*/
/*
 * Unmark: renewed flags and sponsor flags of all key nodes
 */
void KEYTREE_unmark(KEYTREE_control_p_t tree) {
	KEYTREE_unmark_renewed(tree->root);
	KEYTREE_unmark_sponsor(tree);
	KEYTREE_unmark_renewed(tree->subtree->root); 
	KEYTREE_unmark_sponsor(tree->subtree);
}

/*
 * Get the number of members in the tree
 */
int KEYTREE_get_num_of_members(KEYTREE_control_p_t t) {
	KEYTREE_mem_node_p_t ptr = t->m_list_head;
	int num_of_users = 0;

	while (ptr != NULL) {
		num_of_users++;
		ptr = ptr->next;
	}

	return num_of_users;
}

/*
 * Get the members and tree node ids mapping.
 * Return the number of members in the tree
 */
int KEYTREE_get_members_tree_node_ids(KEYTREE_control_p_t t, char** members, 
		int* tree_node_ids) {
	
	KEYTREE_mem_node_p_t m_node;
	int n = 0;	

	m_node = t->m_list_head;
	while (m_node != NULL) {
		memcpy(members[n], m_node->mem_guid, SEAL_GUID_LEN);
		tree_node_ids[n] = m_node->tree_node->tree_node_id;
		n++;
		m_node = m_node->next;
	}

	return n;
}

/*
 * Get the keypath length
 */
int KEYTREE_get_keypath_len(KEYTREE_control_p_t t, char const* guid) {
	KEYTREE_mem_node_p_t m_node;
	KEYTREE_tree_node_p_t k_node;
	int n;	

	if (t->root == NULL) {
		return 0;
	}

	// first locate the m_node which matches the guid
	m_node = t->m_list_head;
	while (m_node != NULL) {
		if (memcmp(m_node->mem_guid, guid, SEAL_GUID_LEN) == 0) {
			break;
		} else {
			m_node = m_node->next;
		}
	}
	if (m_node == NULL) {	// when the keytree is just created before update
		return 0;
	}

	// find the length of the tree path
	n = 0;
	k_node = m_node->tree_node;
	while (k_node != NULL) {
		n++;
		k_node = k_node->parent;
	}
	return n;
}

/*
 * get the key-path (in terms of tree_node_ids) of the member given
 * the member's guid
 *
 * return the size of the keypath
 */
int KEYTREE_get_keypath(KEYTREE_control_p_t t, char const* guid, int* keypath) {
	
	KEYTREE_mem_node_p_t m_node;
	KEYTREE_tree_node_p_t k_node;
	int n;	

	// first locate the m_node which matches the guid
	m_node = t->m_list_head;
	while (m_node != NULL) {
		if (memcmp(m_node->mem_guid, guid, SEAL_GUID_LEN) == 0) {
			break;
		} else {
			m_node = m_node->next;
		}
	}
	if (m_node == NULL) {
		debug_err("member not find in get_keypath\n");
		KEYTREE_print(t->root);
		exit(-1);
	}

	// get the tree node ids of the key path
	n = 0;
	k_node = m_node->tree_node;
	while (k_node != NULL) {
		keypath[n++] = k_node->tree_node_id;
		k_node = k_node->parent;
	}
	return n;
}

/*
 * Get all renewed nodes in the keypath. Put the tree node ids into the
 * list. Return the number of renewed nodes.
 */
int KEYTREE_get_renewed(KEYTREE_control_p_t t, char const* guid, int* renewed) {
	KEYTREE_mem_node_p_t m_node;
	KEYTREE_tree_node_p_t k_node;
	int n;	

	// first locate the m_node which matches the guid
	m_node = t->m_list_head;
	while (m_node != NULL) {
		if (memcmp(m_node->mem_guid, guid, SEAL_GUID_LEN) == 0) {
			break;
		} else {
			m_node = m_node->next;
		}
	}
	if (m_node == NULL) {
		debug_err("member not find in get_renewed\n");
		KEYTREE_print(t->root);
		exit(-1);
	}

	// check if a node is renewed 
	n = 0;
	k_node = m_node->tree_node;
	while (k_node != NULL) {
		if (k_node->renewed) {
			renewed[n] = k_node->tree_node_id;
			n++;
		}	
		k_node = k_node->parent;
	}
	return n;
}

/*
 * Check if the keytree contains no members. Return 1 if yes, 0 if no.
 */
int KEYTREE_is_empty(KEYTREE_control_p_t tree) {
	if (tree->root == NULL) {
		return 1;
	}
	return 0;
}

/*
 * Check if the guid exists in the keytree 
 */
int KEYTREE_is_guid_exist(KEYTREE_control_p_t tree, char const* guid) {

	KEYTREE_mem_node_p_t m_node;

	// first locate the m_node which matches the guid
	m_node = tree->m_list_head;
	while (m_node != NULL) {
		if (memcmp(m_node->mem_guid, guid, SEAL_GUID_LEN) == 0) {
			return 1;
		}
		m_node = m_node->next;
	}	
	return 0;
}


/*----------------------------------------------------------------------
 *                   DEBUG FUNCTIONS
 *----------------------------------------------------------------------*/
void KEYTREE_print(KEYTREE_tree_node_p_t node) {
	static int depth = 0;
	int i;

	if (node == NULL)
		return;

	if (node->right != NULL) {
		depth++;
		KEYTREE_print(node->right);
		depth--;
	}
	for (i=0; i < depth; i++)
		printf("\t");

	printf("%d", node->tree_node_id);
	if (node->renewed == 1) {
		printf("*");
	} 
	if (node->mem_node != NULL) {
		printf("(%s)", node->mem_node->mem_guid);
		if (node->mem_node->sponsor) {
			printf("(s)");
		}
	}
	printf("\n");

	if (node->left != NULL) {
		depth++;
		KEYTREE_print(node->left);
		depth--;
	}
}

/*-------------------------------------------------------------------------
 *					DESTROY
 *-------------------------------------------------------------------------*/
void KEYTREE_destroy(KEYTREE_control_p_t tree) {
	KEYTREE_tree_node_p_t k_node;
	KEYTREE_control_p_t left_subtree, right_subtree;
	
	if (tree == NULL)
		return;

	if (tree->root == NULL) {
		free(tree);
		return;
	}

	KEYTREE_destroy(tree->subtree);

	k_node = tree->root;
	if (k_node->left == NULL && k_node->right == NULL) { 
									// root node = leaf node
		free(k_node->mem_node);
		free(k_node);
		free(tree);
	} else {		// destroy right subtree and left subtree
		left_subtree = KEYTREE_init();
		right_subtree = KEYTREE_init();
		left_subtree->root = k_node->left;
		right_subtree->root = k_node->right;
		free(k_node);
		free(tree);
		KEYTREE_destroy(left_subtree);
		KEYTREE_destroy(right_subtree);
	}
}


/*----------------------------------------------------------------------
 *						STATIC FUNCTIONS
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 * STATIC: BASIC OPERATIONS ON THE KEYTREE
 *----------------------------------------------------------------------*/
/*
 * BATCH: Single join (i.e., join_size == 1)
 * (It gives a simpler operation than pure_join with size == 1
 */
static void KEYTREE_single_join(KEYTREE_control_p_t tree, 
		char const* mem_guid) {
	
	KEYTREE_tree_node_p_t k_new_node;	// key tree node for new join
	KEYTREE_tree_node_p_t k_ins_node;	// insertion tree node	
	KEYTREE_tree_node_p_t tmp_k_node;	// temp k_node
	KEYTREE_list_p_t list;				// list holding temporary tree node

	int height_t;		// height of the existing tree 
	int height_j;		// height the subtree of joined users
	int node_height; 	// height of the subtree rooted at the node
	int node_level;		// level of the node (root level = 0)
	int is_ins_found = 0;	// check if the insertion node is found
	
	// malloc the new tree node (or key node) and corresponding member node
	k_new_node = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
	k_new_node->mem_node = (KEYTREE_mem_node_p_t)
		malloc(sizeof(KEYTREE_mem_node_t));
	k_new_node->left = NULL;
	k_new_node->right = NULL;
	k_new_node->parent = NULL;
	k_new_node->tree_node_id = -1;
	k_new_node->renewed = 0;
	k_new_node->mem_node->tree_node = k_new_node;
	k_new_node->mem_node->prev = NULL;
	k_new_node->mem_node->next = NULL;
	memcpy(k_new_node->mem_node->mem_guid, mem_guid, SEAL_GUID_LEN);
	k_new_node->mem_node->sponsor = 0;

	if (tree->root == NULL) {	// existing tree is empty
		
		// set the tree equal to the new tree node
		tree->root = k_new_node;
		tree->m_list_head = k_new_node->mem_node;
		tree->m_list_tail = k_new_node->mem_node;

		// update the new tree node's id
		k_new_node->tree_node_id = 0;

	} else {				// existing tree has members

		// find the insertion node
		height_t = KEYTREE_get_height(tree->root);	
		height_j = 1;
		list = KEYTREE_list_init();
		KEYTREE_list_addq(list, tree->root);
		k_ins_node = KEYTREE_list_delq(list);
		while (!is_ins_found && k_ins_node != NULL) {
			node_height = KEYTREE_get_height(k_ins_node);
			node_level = log_2(k_ins_node->tree_node_id+1) ;
			if ((node_height > height_j && node_level + 1 + node_height
						<= height_t)
					|| (height_j >= node_height && node_level + 1 + height_j
						<= height_t)) {
				is_ins_found = 1;
			} else {
				if (k_ins_node->left != NULL 
						&& k_ins_node->right != NULL) {
					KEYTREE_list_addq(list, k_ins_node->left);
					KEYTREE_list_addq(list, k_ins_node->right);
				}
				k_ins_node = KEYTREE_list_delq(list);
			}
		}
		if (is_ins_found == 0) {  // inserted at root
			k_ins_node = tree->root;
		}
		KEYTREE_list_destroy(list); 

		// attach the subtree to the insertion node :
		// Create a new node that replaces the insertion node 
		tmp_k_node = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
		tmp_k_node->parent = k_ins_node->parent;
		tmp_k_node->left = k_new_node;
		tmp_k_node->right = k_ins_node;
		tmp_k_node->mem_node = NULL;
		tmp_k_node->tree_node_id = k_ins_node->tree_node_id;
		// Update insertion node's parent (or tmp_k_node's parent) 
		if (tmp_k_node->parent == NULL) {
			tree->root = tmp_k_node;
		} else if (tmp_k_node->parent->left == k_ins_node) {
			tmp_k_node->parent->left = tmp_k_node;
		} else {
			tmp_k_node->parent->right = tmp_k_node;
		}
		// Update the insertion node 
		k_ins_node->parent = tmp_k_node;
		KEYTREE_update_tree_node_id(k_ins_node, 2*tmp_k_node->tree_node_id+2);
		// Update the new node
		k_new_node->parent = tmp_k_node;
		k_new_node->tree_node_id = 2 * tmp_k_node->tree_node_id + 1;

		// Mark all subtree up the key path as "renewed", 
		KEYTREE_mark_renewed(k_new_node->parent);

		// add the member list of the subtree to that of the main tree 
		// (users exist in the main tree) 
		tree->m_list_tail->next = k_new_node->mem_node;
		k_new_node->mem_node->prev = tree->m_list_tail;
		tree->m_list_tail = k_new_node->mem_node;
	
	}	// end: existing tree has members	
}


/*
 * BATCH: Single leave (i.e., leave_size == 1)
 */
static void KEYTREE_single_leave(KEYTREE_control_p_t tree, 
		char const* mem_guid) {

	KEYTREE_tree_node_p_t k_leave_node, k_leave_sib_node, k_leave_par_node;
	KEYTREE_mem_node_p_t m_leave_node;

	// Find out the leaving k-node and its parent 
	m_leave_node = tree->m_list_head;
	while (m_leave_node != NULL && memcmp(m_leave_node->mem_guid, mem_guid,
				SEAL_GUID_LEN) != 0) {
		m_leave_node = m_leave_node->next;
	}

	if (m_leave_node == NULL) {
		debug_err("member %s not found single leave\n", mem_guid);
		KEYTREE_print(tree->root);
		exit(-1);
	}
	k_leave_node = m_leave_node->tree_node;
	k_leave_par_node = k_leave_node->parent;

	// For the case of single existing user 
	if (k_leave_par_node == NULL) {
		tree->root = NULL;
		tree->m_list_head = NULL;
		tree->m_list_tail = NULL;
		free(m_leave_node);
		free(k_leave_node);
		return;
	} else {
		// For the case of multiple users, find the sibling node 
		if (k_leave_par_node->left == k_leave_node) {
			k_leave_sib_node = k_leave_par_node->right;
		} else {
			k_leave_sib_node = k_leave_par_node->left;
		}

		// push the sibling node up and sibling's new parent 
		k_leave_sib_node->parent = k_leave_par_node->parent;
		if (k_leave_sib_node->parent == NULL) {
			tree->root = k_leave_sib_node;
		} else if (k_leave_sib_node->parent->left == k_leave_par_node) {
			k_leave_sib_node->parent->left = k_leave_sib_node;	
		} else {
			k_leave_sib_node->parent->right = k_leave_sib_node;
		}

		// update tree node id's of the subtree rooted at the sibling node 
		KEYTREE_update_tree_node_id(k_leave_sib_node, 
				(k_leave_sib_node->tree_node_id-1)/2);

		// Remove the user from the user_list 
		if (m_leave_node->prev != NULL) {
			m_leave_node->prev->next = m_leave_node->next;
		} else {
			tree->m_list_head = m_leave_node->next;
		}
		if (m_leave_node->next != NULL) {
			m_leave_node->next->prev = m_leave_node->prev;
		} else {
			tree->m_list_tail = m_leave_node->prev;
		}

		// free all user nodes 
		free(m_leave_node);
		free(k_leave_node);
		free(k_leave_par_node);

		// mark all renewed nodes along key path as "renewed"
		KEYTREE_mark_renewed(k_leave_sib_node->parent);

	} // End of the case when multiple users exist originally  
}


/*
 * BATCH: Pure join (i.e., join_size >= 1)
 */
static void KEYTREE_pure_join(KEYTREE_control_p_t tree, char* const* joins, 
		int join_size, KEYTREE_tree_node_p_t insertion) {
	KEYTREE_tree_node_p_t k_ins_node; 	// insertion k-node 
	KEYTREE_tree_node_p_t k_node;
	KEYTREE_list_p_t list;
	int is_ins_found = 0;
	
	KEYTREE_control_p_t subtree;		// subtree of joined users 

	int height_t;		// height of the existing tree 
	int height_j;		// height the subtree of joined users
	int node_height; 	// height of the subtree rooted at the node
	int node_level;		// level of the node (root level = 0)
		
	if (tree->root == NULL) {	// no user exists 
		// create the subtree of all joins 
		subtree = KEYTREE_init();	
		KEYTREE_create_from_size(subtree, joins, join_size);
		
		// set the tree equal to the subtree 
		tree->root = subtree->root;
		tree->m_list_head = subtree->m_list_head;
		tree->m_list_tail = subtree->m_list_tail;
		
		// free unnecessary variables 
		free(subtree);

	} else {	// some users exist originally 
		// Without specified, use BFS to find the insetion node, 
		// Take the node such that the merge won't increase the height 
		if (insertion != NULL) {
			k_ins_node = insertion;
		} else {

			height_t = KEYTREE_get_height(tree->root);	
			height_j = log_2(2*join_size-1) + 1;
			list = KEYTREE_list_init();
			KEYTREE_list_addq(list, tree->root);
			k_ins_node = KEYTREE_list_delq(list);
			while (!is_ins_found && k_ins_node != NULL) {
				node_height = KEYTREE_get_height(k_ins_node);
				node_level = log_2(k_ins_node->tree_node_id+1) ;
				if ((node_height > height_j && node_level + 1 + node_height
							<= height_t)
					|| (height_j >= node_height && node_level + 1 + height_j
						<= height_t)) {
					is_ins_found = 1;
				} else {
					if (k_ins_node->left != NULL 
							&& k_ins_node->right != NULL) {
						KEYTREE_list_addq(list, k_ins_node->left);
						KEYTREE_list_addq(list, k_ins_node->right);
					}
					k_ins_node = KEYTREE_list_delq(list);
				}
			}
			if (is_ins_found == 0) {  // inserted at root
				k_ins_node = tree->root;
			}
			KEYTREE_list_destroy(list); 
		}
		// create a subtree of all joined users 
		subtree = KEYTREE_init();
		KEYTREE_create_from_size(subtree, joins, join_size);
		KEYTREE_update_tree_node_id(subtree->root, 
				2 * k_ins_node->tree_node_id + 1);

		// Add the subtree to the insertion node :
		// Create a new node that replaces the insertion node 
		k_node = (KEYTREE_tree_node_p_t)malloc(sizeof(KEYTREE_tree_node_t));
		k_node->parent = k_ins_node->parent;
		k_node->left = subtree->root;
		k_node->right = k_ins_node;
		k_node->mem_node = NULL;
		k_node->tree_node_id = k_ins_node->tree_node_id;
		// Update insertion node's parent (or k_node's parent) 
		if (k_node->parent == NULL) {
			tree->root = k_node;
		} else if (k_node->parent->left == k_ins_node) {
			k_node->parent->left = k_node;
		} else {
			k_node->parent->right = k_node;
		}
		// Update the insertion node 
		k_ins_node->parent = k_node;
		KEYTREE_update_tree_node_id(k_ins_node, 
				2 * k_node->tree_node_id + 2);
		// Update the subtree's root 
		subtree->root->parent = k_node;
		
		// Mark all subtree up the key path as "renewed", 
		KEYTREE_mark_renewed(subtree->root->parent);

		// add the member list of the subtree to that of the main tree 
		// (users exist in the main tree) 
		tree->m_list_tail->next = subtree->m_list_head;
		subtree->m_list_head->prev = tree->m_list_tail;
		tree->m_list_tail = subtree->m_list_tail;
		
		// free unnecessary variables 
		free(subtree);
	} // End: some members exist originally 
}


/*
 * BATCH: Pure leave (i.e., leave_size >= 1)
 */
static void KEYTREE_pure_leave(KEYTREE_control_p_t tree, char* const* leaves, 
		int leave_size) {
	int i;

	for (i=0; i<leave_size; ++i)
		KEYTREE_single_leave(tree, leaves[i]);
}

/*------------------------------------------------------------------------
 * STATIC: HELPER FUNCTIONS
 *------------------------------------------------------------------------*/
/*
 * Update the tree_node_id's of all nodes in a subtree.
 */
static void KEYTREE_update_tree_node_id(KEYTREE_tree_node_p_t root, 
		int root_id) {
	if (root != NULL) {
		root->tree_node_id = root_id;
	}
	if (root->left != NULL) {
		KEYTREE_update_tree_node_id(root->left, 2 * root_id + 1);	
	}	
	if (root->right != NULL) {
		KEYTREE_update_tree_node_id(root->right, 2 * root_id + 2);
	}
}

/*
 * Mark nodes, including the node itself and its ancestor, 
 * along the key path as "renewed"
 */
static void KEYTREE_mark_renewed(KEYTREE_tree_node_p_t k_node) {
	while (k_node != NULL) {
		k_node->renewed = 1;
		k_node = k_node->parent;
	}
}
		
/*
 * Unmark all renewed nodes
 */
static void KEYTREE_unmark_renewed(KEYTREE_tree_node_p_t k_node) {
	
	if (k_node != NULL) {
		k_node->renewed = 0;
		KEYTREE_unmark_renewed(k_node->left);
		KEYTREE_unmark_renewed(k_node->right);
	}
}

/*
 * Find and mark the sponsor.
 */
static void KEYTREE_mark_sponsor(KEYTREE_tree_node_p_t k_node) {
	while (!(k_node->left == NULL && k_node->right == NULL)) {
		k_node = k_node->right;
	}
	if (k_node->mem_node == NULL) {
		debug("user node not found when finding sponsor.\n");
		return;
	} else {
		k_node->mem_node->sponsor = 1;
	}
}

/*
 * Unmark all sponsors 
 */
static void KEYTREE_unmark_sponsor(KEYTREE_control_p_t tree) {
	KEYTREE_mem_node_p_t u_node;

	u_node = tree->m_list_head;
	while (u_node != NULL) {
		u_node->sponsor = 0;
		u_node = u_node->next;
	}
}

/*
 * Mark all sponsors 
 */
static void KEYTREE_mark_all_sponsor(KEYTREE_tree_node_p_t k_node) {
	if (k_node != NULL && k_node->renewed) {
		if (k_node->left != NULL) {
			if (k_node->left->renewed) {
				KEYTREE_mark_all_sponsor(k_node->left);
			} else {
				KEYTREE_mark_sponsor(k_node->left);
			}
		}
		if (k_node->right != NULL) {
			if (k_node->right->renewed) {
				KEYTREE_mark_all_sponsor(k_node->right);
			} else {
				KEYTREE_mark_sponsor(k_node->right);
			}
		}
	}
}

/*
 * Get the tree height
 */
static int KEYTREE_get_height(KEYTREE_tree_node_p_t node) {
	int height;
	int left_tree_height = 0;
	int right_tree_height = 0;

	if (node->left == NULL && node->right == NULL) {
		height = 1;
	} else {
		left_tree_height = KEYTREE_get_height(node->left);
		right_tree_height = KEYTREE_get_height(node->right);

		if (left_tree_height > right_tree_height) {
			height = left_tree_height + 1;
		} else {
			height = right_tree_height + 1;
		}
	}
	return height;
}

/*--------------------------------------------------------------------
 *	STATIC:	LINKED LIST HANDLING
 *--------------------------------------------------------------------*/
/*
 * Create and initialize a double linked list
 */
static KEYTREE_list_p_t KEYTREE_list_init() {
	KEYTREE_list_p_t list;	

	list = (KEYTREE_list_p_t)malloc(sizeof(KEYTREE_list_t));

	list->head = NULL;
	list->tail = NULL;

	return list;
}

/*
 * Add a new node: Enqueue 
 */
static void KEYTREE_list_addq(KEYTREE_list_p_t list_p, 
		KEYTREE_tree_node_p_t tnode) {

	KEYTREE_list_node_p_t new_node_p;

	new_node_p = (KEYTREE_list_node_p_t)malloc(sizeof(KEYTREE_list_node_t));
	new_node_p->tnode = tnode;
	if (list_p->head == NULL) {
		new_node_p->prev = NULL;
		new_node_p->next = NULL;
		list_p->head = new_node_p;
		list_p->tail = new_node_p;
	} else {
		new_node_p->prev = list_p->tail;
		new_node_p->next = NULL;
		list_p->tail->next = new_node_p;
		list_p->tail = new_node_p;
	}
}

/*
 * Delete an item: Dequeue
 */
static KEYTREE_tree_node_p_t KEYTREE_list_delq(KEYTREE_list_p_t list_p) {
	KEYTREE_list_node_p_t del_node_p;
	KEYTREE_tree_node_p_t ret_tnode = NULL;

	if (list_p->head == NULL) {
		return NULL;
	}

	del_node_p = list_p->head;
	ret_tnode = del_node_p->tnode;

	if (list_p->head == list_p->tail) {
		list_p->head = NULL;
		list_p->tail = NULL;
	} else {
		list_p->head = del_node_p->next;
		list_p->head->prev = NULL;
	}

	free(del_node_p);
	return ret_tnode;
}

/*
 * Add the item: in increasing order of tree node id
 */
void KEYTREE_list_addi(KEYTREE_list_p_t list_p, KEYTREE_tree_node_p_t tnode) {

	KEYTREE_list_node_p_t new_node_p = NULL, tmp_node_p = NULL;

	// malloc and initialize the new node	
	new_node_p = (KEYTREE_list_node_p_t)malloc(sizeof(KEYTREE_list_node_t));
	new_node_p->tnode = tnode;

	// search and add the new node in correct position
	if (list_p->head != NULL && list_p->tail != NULL) {
		tmp_node_p = list_p->head;
		while (tmp_node_p != NULL) {
			if (new_node_p->tnode->tree_node_id 
					== tmp_node_p->tnode->tree_node_id) {
				break;
			} else if (new_node_p->tnode->tree_node_id 
					< tmp_node_p->tnode->tree_node_id)  {
				if (tmp_node_p == list_p->head) {      // add at the head
					new_node_p->prev = NULL;
					new_node_p->next = tmp_node_p;
					tmp_node_p->prev = new_node_p;
					list_p->head = new_node_p;
				} else {                // add in the middle
					new_node_p->prev = tmp_node_p->prev;
					new_node_p->next = tmp_node_p;
					tmp_node_p->prev->next = new_node_p;
					tmp_node_p->prev = new_node_p;
				}
				break;
			}
			tmp_node_p = tmp_node_p->next;
		}
		if (tmp_node_p == NULL) {     // add at the end
			new_node_p->prev = list_p->tail;
			new_node_p->next = NULL;
			list_p->tail->next = new_node_p;
			list_p->tail = new_node_p;
		}
	} else {            // the list is empty
		new_node_p->prev = NULL;
		new_node_p->next = NULL;
		list_p->head = new_node_p;
		list_p->tail = new_node_p;
	}
}

/*
 * Delete an item: Pop
 */
static KEYTREE_tree_node_p_t KEYTREE_list_pop(KEYTREE_list_p_t list_p) {
	KEYTREE_tree_node_p_t ret_tnode = NULL;
	KEYTREE_list_node_p_t del_node_p;

	if (list_p->head == NULL) {
		return NULL;
	}

	del_node_p = list_p->tail;
	ret_tnode = del_node_p->tnode;

	if (list_p->head == list_p->tail) {
		list_p->head = NULL;
		list_p->tail = NULL;
	} else {
		list_p->tail = del_node_p->prev;
		list_p->tail->next = NULL;
	}

	free(del_node_p);
	return ret_tnode;
}

/*
 * Destroy the whole list
 */
static void KEYTREE_list_destroy(KEYTREE_list_p_t list_p) {

	KEYTREE_tree_node_p_t ret_node_p = NULL;;

	ret_node_p = KEYTREE_list_delq(list_p);
	while (ret_node_p != NULL) {
		ret_node_p = KEYTREE_list_delq(list_p);
	}
	free(list_p);
}

/*
 * Get the truncated value of log_2(x)
 */
static int log_2(int x) {
	int mask = 0xFFFFFFFF;
	int value = x >> 1;
	int result = 0;

	while ((value & mask) != 0) {
		result++;
		value = value >> 1;
	}
	return result;
}


