/*
 * leader_engine.c
 *
 * - Leader management, including election. 
 * - The leader is the one who is responsible to broadcast rekey message. 
 *   He's the one who has stayed for the longest time in the group.
 */

#include "seal.h"

/*
 * Initialize the leader control sturcture 
 */
LEADER_control_p_t LEADER_init() {

	LEADER_control_p_t leader;

	leader = (LEADER_control_p_t)malloc(sizeof(LEADER_control_t));
	leader->leader_flag = -1;
	memset(leader->leader_guid, 0, SEAL_GUID_LEN);
	leader->leader_seq = 0;
	pthread_mutex_init(&(leader->lock), NULL);

	return leader;
}

/*------------------------------------------------------------------------
 *                        LEADER MANAGEMENT
 *------------------------------------------------------------------------*/
/*
 * Return 1 if I am the leader, and 0 otherwise
 */
int LEADER_elect(LEADER_control_p_t leader, char const* my_guid, int no_curr,
		char* const* curr_guids, int const* curr_ts) {

	int isLeader = 0;	
	int i;

	// check if I am the leader, i.e., I am the only one who have the
	// timestamp = 0 in curr_memlist_g
	for (i=0; i<no_curr; ++i) {
		if (curr_ts[i] == 0) {		// who have previously join
			if (memcmp(curr_guids[i], my_guid, SEAL_GUID_LEN) == 0) {
				isLeader = 1;
			} else {		
				isLeader = 0;
				break;
			}
		}
	}

	// debug
	if (isLeader) {					// I am the leader
		debug("I am the leader: %s\n", my_guid);
	}	

	return isLeader;
}

/*
 * Set the leader
 */
void LEADER_set_leader(LEADER_control_p_t leader, char const* my_guid, 
		char const* leader_guid) {
	if (leader_guid == 0) {
		memset(leader->leader_guid, 0, SEAL_GUID_LEN);
		leader->leader_flag = -1;
	} else if (memcmp(leader_guid, my_guid, SEAL_GUID_LEN) != 0) {
		memcpy(leader->leader_guid, leader_guid, SEAL_GUID_LEN);
		leader->leader_flag = 0;
	} else {			// leader_guid == my_guid
		memcpy(leader->leader_guid, leader_guid, SEAL_GUID_LEN);
		leader->leader_flag = 1;
	}
}

/*------------------------------------------------------------------------
 *                        SEQUENCE NUMBER MANAGEMENT
 *------------------------------------------------------------------------*/
/*
 * update the sequence number with the maximum one
 */ 
void LEADER_seq_max(LEADER_control_p_t leader, int recv_seq) {
	if (recv_seq > leader->leader_seq) {
		leader->leader_seq = recv_seq;
	}
	//leader->leader_seq = recv_seq > leader->leader_seq ?  recv_seq : 
	//	leader->leader_seq;
}

/*
 * increment the sequence number 
 */
void LEADER_seq_increment(LEADER_control_p_t leader) {
	leader->leader_seq++;
}

/*
 * retrieve the sequence number
 */
int LEADER_seq_retrieve(LEADER_control_p_t leader) {
	return leader->leader_seq;
}

/*------------------------------------------------------------------------
 *                        CHECKING FUNCTIONS
 *------------------------------------------------------------------------*/
/*
 * Check if leader is changed. 1 - yes, 0 - no
 */
int LEADER_is_changed(LEADER_control_p_t leader) {
	if (leader->leader_flag == -1) {
		return 1;
	}
	return 0;
}

/*
 * Determine if election is necessary. 1 - yes, 0 - no
 * Require election if (1) I am the newly joining member, (2) I am uncertain
 * who is leader or the leader leaves.
 */
int LEADER_is_elect(LEADER_control_p_t leader, char const* my_guid, 
		char const* mem_guid, int event) {
	
	if (event == LEADER_JOIN_EVENT) {
		if (memcmp(mem_guid, my_guid, SEAL_GUID_LEN) == 0) {
			return 1;
		} else {
			return 0;
		}	
	} else if (event == LEADER_LEAVE_EVENT) {
		if (leader->leader_flag == -1 || (leader->leader_flag == 0 && 
					memcmp(mem_guid, leader->leader_guid, SEAL_GUID_LEN) 
					== 0)) {
			return 1;
		} else {
			return 0;
		}	
	}

	// undefined type (should never reach here)
	return 0;
}

/*
 * Check if I am the leader. 1 - yes, 0 - no
 */
int LEADER_is_leader(LEADER_control_p_t leader) {
	if (leader->leader_flag == 1) {
		return 1;
	}
	return 0;
}

/*
 * Check if the "leader" is a valid leader.
 * Return 1 if the leader is valid, 0 otherwise.
 */
int LEADER_is_valid(LEADER_control_p_t leader, char const* claimed_leader) {
	int ret = 1;

	// if a dishonest member claims himself to be the leader, the "true"
	// leader and other certain members will detect the fraud.

	if (leader->leader_flag == 1 || leader->leader_flag == 0) {		
		// I am the leader or I am not the leader but I know who the leader is
		if (memcmp(claimed_leader, leader->leader_guid, SEAL_GUID_LEN) != 0) {
			// invalid
			ret = 0;
		}
	}

	return ret;
}


/* 
 * Destroy
 */
void LEADER_destroy(LEADER_control_p_t leader) {
	free(leader);
}
