/*
 * rekey_poll_thread.c
 *
 * - polling rekey messages periodically
 */

#include "seal.h"

static RQUEUE_control_p_t rekey_queue_g;

/*
 * the thread function
 */
void* REKEY_POLL_init(SEAL_session_p_t seal) {
	pthread_cond_t cond;        // dummy condition variable
	pthread_mutex_t lock;       // dummy lock

	struct timeval tv;          // time structures
	struct timespec ts;

	int i;                      // loop counter

	int no_join = -1, no_leave = -1;
	char **join_guids = NULL, **leave_guids = NULL;
	
	// set cancel state
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_cleanup_push((void*)&REKEY_POLL_cleanup_handler, NULL);

	// set static variables
	rekey_queue_g = seal->rekey_queue;

	// initialize the variables
	pthread_cond_init(&cond, NULL);
	pthread_mutex_init(&lock, NULL);
	pthread_mutex_lock(&lock);
	gettimeofday(&tv, NULL);        // get the current time
	ts.tv_sec = tv.tv_sec;
	ts.tv_nsec = tv.tv_usec * 1000;

	while (1) {
		// acquire all locks    
		pthread_mutex_lock(&(rekey_queue_g->lock));

		// enqueue a rekey message
		// (determine the rekeying version)
#ifdef REBUILD
		RQUEUE_enq(rekey_queue_g, REKEY_REBUILD, NULL, NULL);
#else
#ifdef BATCH
		RQUEUE_enq(rekey_queue_g, REKEY_BATCH, NULL, NULL);
#else
#ifdef QBATCH
		RQUEUE_enq(rekey_queue_g, REKEY_QMERGE, NULL, NULL);
#endif
#endif
#endif

		pthread_cond_broadcast(&(rekey_queue_g->empty_cond));

		// release the lock
		pthread_mutex_unlock(&(rekey_queue_g->lock));
		
		// free all resources
		for (i=0; i<no_join; ++i)
			free(join_guids[i]);
		for (i=0; i<no_leave; ++i)
			free(leave_guids[i]);
		free(join_guids);
		free(leave_guids);

		// set the next running schedule
		ts.tv_sec = ts.tv_sec + REKEY_POLL_PERIOD;
		pthread_cond_timedwait(&cond, &lock, &ts);
	}

	pthread_cleanup_pop(1);

	return NULL;
}

/*
 * cleanup handler
 */
void* REKEY_POLL_cleanup_handler() {
	pthread_mutex_trylock(&(rekey_queue_g->lock));
	pthread_mutex_unlock(&(rekey_queue_g->lock));
	debug("REKEY POLL->cleanup\n");
	return NULL;
}
